"""Tests for the Meta class."""
from sprynger import init
from sprynger import Meta
from sprynger.utils.data_structures import (
    MetadataCreator,
    MetaDiscipline,
    MetadataFacets,
    MetaRecord,
    MetaURL,
)

init()

article = Meta(doi='10.1007/s00394-024-03496-7', refresh=True)

def test_results():
    """Test the results."""
    assert article.results.total == 1
    assert article.results.start == 1
    assert article.results.pageLength == 10
    assert article.results.recordsRetrieved == 1

def test_records():
    """Test the records."""
    expected_record = MetaRecord(
        contentType="Article",
        identifier="doi:10.1007/s00394-024-03496-7",
        language="en",
        urls=[
            MetaURL(
                format="html",
                platform="web",
                value="http://link.springer.com/openurl/fulltext?id=doi:10.1007/s00394-024-03496-7",
            ),
            MetaURL(
                format="pdf",
                platform="web",
                value="http://link.springer.com/openurl/pdf?id=doi:10.1007/s00394-024-03496-7",
            ),
            MetaURL(
                format="",
                platform="",
                value="http://dx.doi.org/10.1007/s00394-024-03496-7",
            ),
        ],
        title="Ultra-processed food intake in toddlerhood and mid-childhood in the UK: cross sectional and longitudinal perspectives",
        creators=[
            MetadataCreator(creator="Conway, Rana E.", ORCID="0000-0003-0955-7107"),
            MetadataCreator(creator="Heuchan, Gabriella N.", ORCID=None),
            MetadataCreator(creator="Heggie, Lisa", ORCID="0000-0002-4846-2357"),
            MetadataCreator(creator="Rauber, Fernanda", ORCID="0000-0001-9693-7954"),
            MetadataCreator(creator="Lowry, Natalie", ORCID="0000-0002-9137-5005"),
            MetadataCreator(creator="Hallen, Hannah", ORCID=None),
            MetadataCreator(creator="Llewellyn, Clare H.", ORCID="0000-0002-0066-2827"),
        ],
        publicationName="European Journal of Nutrition",
        openaccess=True,
        doi="10.1007/s00394-024-03496-7",
        publisher="Springer",
        publicationDate="2024-12-01",
        publicationType="Journal",
        issn="1436-6207",
        eIssn="1436-6215",
        volume=63,
        number=8,
        issueType="Regular",
        topicalCollection="",
        genre=["OriginalPaper", "Original Contribution"],
        startingPage=3149,
        endingPage=3160,
        journalId=394,
        onlineDate="2024-10-04",
        copyright="©2024 The Author(s)",
        abstract="Purpose (i) Characterize ultra-processed food (UPF) intakes in toddlerhood and mid-childhood, including identifying principal UPF sub-groups and associations with nutrient profile; (ii) explore stability and change in UPF intake between toddlerhood and mid-childhood. Methods Data were from children in the UK Gemini twin cohort at 21 months ( n \u2009=\u20092,591) and 7 years ( n \u2009=\u2009592) of age. UPF intakes were estimated using diet diaries and Nova classification. Complex samples general linear or logistic regression models were used to explore associations between UPF intake, UPF sub-groups and nutrients, and changes in intake over time. Results The contribution of UPF to total energy was 46.9% (±\u200914.7) at 21 months and 59.4% (±\u200912.5) at 7 years. Principal UPF sub-groups were yogurts, higher-fiber breakfast cereals, and wholegrain breads in toddlerhood, and puddings and sweet cereal products and white breads in mid-childhood. At both ages, mean free sugar and sodium intakes exceeded recommended maximums and higher UPF consumption was associated with consuming more of each nutrient ( P \u2009<\u20090.001). UPF intake was negatively associated with fat, saturated fat and protein intake in toddlerhood, and fiber intake in mid-childhood ( P \u2009<\u20090.001). Being in the highest UPF intake quintile in toddlerhood was predictive of being in the highest quintile in mid-childhood (OR 9.40, 95%CI 3.94–22.46). Conclusions UPF accounted for nearly half of toddlers’ energy, increasing to 59% in mid-childhood. Higher UPF consumers had higher intakes of free sugar and sodium. UPF intake in toddlerhood was predictive of mid-childhood intake. Effective policies are needed to reduce UPF intakes in the early years.",
        conferenceInfo=[],
        keyword=["Ultra-processed foods", "Diet quality", "Toddlers", "Children", "UK"],
        subjects=["Chemistry", "Nutrition"],
        disciplines=[MetaDiscipline(id="3524", term="Nutrition")],
    )
    assert article.records[0] == expected_record

def test_facets():
    """Test the facets."""
    assert len(article.facets) == 12

    expected_first_facet = MetadataFacets(facet='subject', value='Chemistry', count=1)
    assert article.facets[0] == expected_first_facet



"""Module with auxiliary functions to parse OpenAccess documents."""
from typing import Optional

from lxml.etree import _Element

from sprynger.utils.data_structures import Affiliation, Contributor, Date, Paragraph, Reference
from sprynger.utils.parse import get_attr, get_text, make_int_if_possible, stringify_descendants


def affs_to_dict(affs) -> dict[str, Affiliation]:
    """Auxiliary function to query the affiliations by their number."""
    return {aff.nr: aff for aff in affs}


def get_acknowledgements(back: _Element) -> Optional[str]:
    """Parse the acknowledgements of the document."""
    paragraphs = []
    if back is not None:
        for p in back.findall('.//ack/p', []):
            paragraphs.append(get_text(p, '.'))
    return '\n'.join(paragraphs) or None


def get_affiliations(data: _Element) -> list[Affiliation]:
    """Parse the affiliations of the document."""
    affiliations = []
    if data is not None:
        for contrib_group in data.findall('.//contrib-group'):
            contribution_group = contrib_group.get('content-type')
            for a in contrib_group.findall('.//aff'):
                institution = a.find('.//institution-wrap')
                new_aff = Affiliation(
                    type=contribution_group,
                    ref_nr=a.get('id'),
                    ror=get_attr(institution, 'institution-id', 'institution-id-type', 'ROR'),
                    grid=get_attr(institution, 'institution-id', 'institution-id-type', 'GRID'),
                    isni=get_attr(institution, 'institution-id', 'institution-id-type', 'ISNI'),
                    division=get_attr(institution, 'institution', 'content-type', 'org-division'),
                    name=get_attr(institution, 'institution', 'content-type', 'org-name'),
                    city=get_attr(a, 'addr-line', 'content-type', 'city'),
                    country=get_text(a, './/country')
                )
                affiliations.append(new_aff)
    return affiliations


def get_contributors(data: _Element) -> list[Contributor]:
    """Parse the contributors of the document and matcg them with their affiliations."""
    contributors = []
    if data is not None:
        for c in data.findall('.//contrib'):
            # Get affiliation
            affs_nr = []
            for aff_ref in c.findall('.//xref[@ref-type="aff"]'):
                aff_nr = aff_ref.get('rid')
                affs_nr.append(aff_nr)

            # Get contributor data
            new_contrib = Contributor(
                type=c.get('contrib-type'),
                nr=c.get('id'),
                orcid=get_attr(c, 'contrib-id', 'contrib-id-type', 'orcid'),
                surname=get_text(c, './/name/surname'),
                given_name=get_text(c, './/name/given-names'),
                email=get_text(c, './/email'),
                affiliations_ref_nr=affs_nr
            )
            contributors.append(new_contrib)
    return contributors


def get_date(date_node: _Element) -> Date:
    """Auxiliary function to extract date information from a date node."""
    return Date(day=make_int_if_possible(get_text(date_node, './/day')),
                month=make_int_if_possible(get_text(date_node, './/month')),
                year=make_int_if_possible(get_text(date_node, './/year')))


def get_paragraphs(xml) -> list[Paragraph]:
    """Paragraphs of the OpenAccess document.

    Returns:
        list[Paragraph]: A list of Paragraph objects containing the 
        `paragraph_id`, `section_id`, `section_title`, and `text`.
    """
    if xml is not None:
        parsed_paragraphs = []
        for paragraph in xml.findall('.//p[@id]'):
            paragraph_id = paragraph.get('id')
            paragraph_text = ''.join(paragraph.itertext())

            parent = paragraph.getparent()
            section_id = parent.get('id')
            section_title = get_text(parent, 'title')


            paragraph_data = Paragraph(
                paragraph_id=paragraph_id,
                section_id=section_id,
                section_title=section_title,
                text=paragraph_text.strip()
            )
            parsed_paragraphs.append(paragraph_data)

        return parsed_paragraphs
    return []

def _get_doi(ref_node: _Element) -> Optional[str]:
    """Parse DOIs from a reference node."""
    doi = get_attr(ref_node, "pub-id", "pub-id-type", "doi")
    if doi is None:
        doi = get_attr(ref_node, "ext-link", "ext-link-type", "doi")
        if doi is not None:
            doi = doi.replace("https://doi.org/", "")
    return doi

def _get_names(ref_node: _Element) -> Optional[list[str]]:
    """Parse names from a reference node."""
    names = []
    for person in ref_node.findall('name'):
        given_name = stringify_descendants(person.find('given-names'))
        surname = stringify_descendants(person.find('surname'))
        name = f'{given_name} {surname}'
        names.append(name)
    return names

def _get_names_from_group(ref_node: _Element, person_group_type: str) -> Optional[list[str]]:
    """Parse names from a person-group node."""
    names = []
    group_node = ref_node.find(f'./person-group[@person-group-type="{person_group_type}"]')
    if group_node is not None:
        names = _get_names(group_node)
    return names

def _get_reference(reference: _Element) -> Optional[_Element]:
    """Get reference from one of the three possible positions."""
    for tag in ["mixed-citation", "element-citation", "citation"]:
        ref = reference.find(tag)
        if ref is not None:
            return ref
    return None

def get_reference_list(back: _Element) -> list[Reference]:
    """Parse the references of the document."""
    new_ref_list = []
    if back is not None:
        for ref_list in back.findall('.//ref-list[@id]'):

            ref_list_id = ref_list.get('id')
            ref_list_title = get_text(ref_list, './/title')
            for ref in ref_list.findall('.//ref[@id]'):
                ref_id = ref.get('id')
                ref_label = get_text(ref, 'label')
                ref = _get_reference(ref)
                # Avoid unbound variables
                ref_publication_type, authors, editors, names = None, [], [], []
                ref_title, ref_source, ref_year, ref_doi = None, None, None, None
                if ref is not None:
                    ref_publication_type = ref.get('publication-type') or ref.get('citation-type')
                    if ref_publication_type is not None:
                        # Get the title from the article-title tag or the whole reference
                        ref_title = get_text(ref, "article-title")
                        if ref_title is None:
                            ref_title = stringify_descendants(ref)
                        # Get the authors from the name tag or the person-group tag
                        if ref.find('name') is not None:
                            names = _get_names(ref)
                        else:
                            authors = _get_names_from_group(ref, "author")
                            editors = _get_names_from_group(ref, "editor")
                        # Get the source, year and DOI
                        ref_source = get_text(ref, "source")
                        ref_year = get_text(ref, "year")
                        # Get the DOI
                        ref_doi = _get_doi(ref)

                reference = Reference(
                    ref_list_id=ref_list_id,
                    ref_list_title=ref_list_title,
                    ref_id=ref_id,
                    ref_label=ref_label,
                    ref_publication_type=ref_publication_type,
                    authors=authors,
                    editors=editors,
                    names=names,
                    ref_title=ref_title,
                    ref_source=ref_source,
                    ref_year=ref_year,
                    ref_doi=ref_doi
                )

                new_ref_list.append(reference)
    return new_ref_list



"""Module to initialize the sprynger library."""
import tomllib
import os

from pathlib import Path
from typing import Optional, Union

from sprynger.exceptions import MissingAPIKeyError
from sprynger.utils.constants import DEFAULT_PATHS, REQUESTS

API_KEY = None
CONFIG = None


def init(api_key: Optional[str] = None,
         config_file: Optional[Union[str, Path]] = None) -> None:
    """
    Function to initialize the sprynger library. For more information go to the
    `documentation <file:///Users/nilsherrmann/sprynger/docs/build/html/initialization.html#configuration>`_.
    
    Args:
        api_key (str): API key
        config_file (str): Path to the configuration .toml file.
        
    
    Raises:
        ValueError: If no api key was provided either as argument or as an
        environment variable `API_KEY`.
    
    Example:
        >>> from sprynger import init
        >>> init(key='your key', config_file='path/to/custom/config.toml')
    """
    global API_KEY
    global CONFIG

    CONFIG = _load_default_config()

    if config_file:
        with open(config_file, 'rb') as f:
            custom_config = tomllib.load(f)
        _merge_dicts(CONFIG, custom_config)

    _create_cache_folders(CONFIG)

    API_KEY = api_key or os.environ.get("API_KEY")
    if not API_KEY:
        raise ValueError('No API key found. Provide an API key or set the '
                         'environment variable API_KEY. To get an API key '
                         'visit: https://dev.springernature.com/')


def _load_default_config() -> dict:
    """Auxiliary function to load the default configuration."""
    config = {}
    config['Directories'] = DEFAULT_PATHS
    config['Requests'] = REQUESTS
    return config


def _merge_dicts(default, custom):
    """
    Recursively merge two dictionaries. The values from the custom dictionary
    will overwrite those from the default dictionary.
    """
    for key, value in custom.items():
        if isinstance(value, dict) and key in default:
            _merge_dicts(default[key], value)
        else:
            default[key] = value

def _create_cache_folders(config: dict) -> None:
    """Auxiliary function to create cache folders."""
    directories = config.get('Directories', {})
    for _, path in directories.items():
        cache_path = Path(path)
        cache_path.mkdir(parents=True, exist_ok=True)


def get_config() -> dict:
    """Function to get the config parser."""
    if not CONFIG:
        raise MissingAPIKeyError('Library not initialized. '
                                'Please initialize sprynger with init().\n'
                                'For more information visit: '
                                'the documentation.')
    return CONFIG


def get_key() -> str:
    """Function to get the API keys and overwrite keys in config if needed."""
    return API_KEY



"""Test the startup of the sprynger package."""
from sprynger import init
from sprynger.utils.parse import chained_get
from sprynger.utils.startup import get_config, get_key

def test_custom_api():
    """Test use of custom config file"""
    init(api_key='not existing key',
         config_file='./sprynger/tests/test_config.toml')
    config = get_config()
    assert chained_get(config, ['Directories', 'Metadata']) == './sprynger/tests/'

def test_init():
    """Test the init function."""
    init()
    assert get_config() is not None
    assert get_key() is not None
